from typing import cast

from tortoise import ConfigurationError
from tortoise.backends.mssql.schema_generator import MSSQLSchemaGenerator
from tortoise.fields import UUIDField, TextField, JSONField
from tortoise.indexes import Index


class MsSqlGenerator(MSSQLSchemaGenerator):

    def __init__(self, client) -> None:
        super().__init__(client)
        self.comments_array = []

    def get_table_sql_new(self, model, safe: bool = True) -> dict:
        fields_to_create = []
        fields_with_index = []
        m2m_tables_for_create = []
        references = set()
        models_to_create = []
        if model._meta.db == self.client:
            model._check()
            models_to_create.append(model)
        models_tables = [model._meta.db_table for model in models_to_create]
        for field_name, column_name in model._meta.fields_db_projection.items():
            field_object = model._meta.fields_map[field_name]
            comment = (
                self._column_comment_generator(
                    table=model._meta.db_table, column=column_name, comment=field_object.description
                )
                if field_object.description
                else ""
            )

            default = field_object.default
            auto_now_add = getattr(field_object, "auto_now_add", False)
            auto_now = getattr(field_object, "auto_now", False)
            if default is not None or auto_now or auto_now_add:
                if callable(default) or isinstance(field_object, (UUIDField, TextField, JSONField)):
                    default = ""
                else:
                    default = field_object.to_db_value(default, model)
                    try:
                        default = self._column_default_generator(
                            model._meta.db_table,
                            column_name,
                            self._escape_default_value(default),
                            auto_now_add,
                            auto_now,
                        )
                    except NotImplementedError:
                        default = ""
            else:
                default = ""

            # TODO: PK generation needs to move out of schema generator.
            if field_object.pk:
                if field_object.generated:
                    generated_sql = field_object.get_for_dialect(self.DIALECT, "GENERATED_SQL")
                    if generated_sql:  # pragma: nobranch
                        fields_to_create.append(
                            self.GENERATED_PK_TEMPLATE.format(
                                field_name=column_name,
                                generated_sql=generated_sql,
                                comment=comment,
                            )
                        )
                        continue

            nullable = "NOT NULL" if not field_object.null else ""
            unique = "UNIQUE" if field_object.unique else ""

            if getattr(field_object, "reference", None):
                reference = cast("ForeignKeyFieldInstance", field_object.reference)
                comment = (
                    self._column_comment_generator(
                        table=model._meta.db_table,
                        column=column_name,
                        comment=reference.description,
                    )
                    if reference.description
                    else ""
                )

                to_field_name = reference.to_field_instance.source_field
                if not to_field_name:
                    to_field_name = reference.to_field_instance.model_field_name

                field_creation_string = self._create_string(
                    db_column=column_name,
                    field_type=field_object.get_for_dialect(self.DIALECT, "SQL_TYPE"),
                    nullable=nullable,
                    unique=unique,
                    is_primary_key=field_object.pk,
                    comment="",
                    default=default,
                ) + (
                    self._create_fk_string(
                        constraint_name=self._generate_fk_name(
                            model._meta.db_table,
                            column_name,
                            reference.related_model._meta.db_table,
                            to_field_name,
                        ),
                        db_column=column_name,
                        table=reference.related_model._meta.db_table,
                        field=to_field_name,
                        on_delete=reference.on_delete,
                        comment=comment,
                    )
                    if reference.db_constraint
                    else ""
                )
                references.add(reference.related_model._meta.db_table)
            else:
                field_creation_string = self._create_string(
                    db_column=column_name,
                    field_type=field_object.get_for_dialect(self.DIALECT, "SQL_TYPE"),
                    nullable=nullable,
                    unique=unique,
                    is_primary_key=field_object.pk,
                    comment=comment,
                    default=default,
                )

            fields_to_create.append(field_creation_string)

            if field_object.index and not field_object.pk:
                fields_with_index.append(column_name)

        if model._meta.unique_together:
            for unique_together_list in model._meta.unique_together:
                unique_together_to_create = []

                for field in unique_together_list:
                    field_object = model._meta.fields_map[field]
                    unique_together_to_create.append(field_object.source_field or field)

                fields_to_create.append(
                    self._get_unique_constraint_sql(model, unique_together_to_create)
                )

        # Indexes.
        _indexes = [
            self._get_index_sql(model, [field_name], safe=safe) for field_name in fields_with_index
        ]

        if model._meta.indexes:
            for indexes_list in model._meta.indexes:
                if not isinstance(indexes_list, Index):
                    indexes_to_create = []
                    for field in indexes_list:
                        field_object = model._meta.fields_map[field]
                        indexes_to_create.append(field_object.source_field or field)

                    _indexes.append(self._get_index_sql(model, indexes_to_create, safe=safe))
                else:
                    _indexes.append(indexes_list.get_sql(self, model, safe))

        field_indexes_sqls = [val for val in list(dict.fromkeys(_indexes)) if val]

        fields_to_create.extend(self._get_inner_statements())

        table_fields_string = "\n    {}\n".format(",\n    ".join(fields_to_create))
        table_comment = (
            self._table_comment_generator(
                table=model._meta.db_table, comment=model._meta.table_description
            )
            if model._meta.table_description
            else ""
        )

        table_create_string = self.TABLE_CREATE_TEMPLATE.format(
            exists="IF NOT EXISTS " if safe else "",
            table_name=model._meta.db_table,
            fields=table_fields_string,
            comment=table_comment,
            extra=self._table_generate_extra(table=model._meta.db_table),
        )

        table_create_string = "\n".join([table_create_string, *field_indexes_sqls])

        table_create_string += self._post_table_hook()

        for m2m_field in model._meta.m2m_fields:
            field_object = cast("ManyToManyFieldInstance", model._meta.fields_map[m2m_field])
            if field_object._generated or field_object.through in models_tables:
                continue
            m2m_create_string = self.M2M_TABLE_TEMPLATE.format(
                exists="IF NOT EXISTS " if safe else "",
                table_name=field_object.through,
                backward_fk=self._create_fk_string(
                    "",
                    field_object.backward_key,
                    model._meta.db_table,
                    model._meta.db_pk_column,
                    field_object.on_delete,
                    "",
                )
                if field_object.db_constraint
                else "",
                forward_fk=self._create_fk_string(
                    "",
                    field_object.forward_key,
                    field_object.related_model._meta.db_table,
                    field_object.related_model._meta.db_pk_column,
                    field_object.on_delete,
                    "",
                )
                if field_object.db_constraint
                else "",
                backward_key=field_object.backward_key,
                backward_type=model._meta.pk.get_for_dialect(self.DIALECT, "SQL_TYPE"),
                forward_key=field_object.forward_key,
                forward_type=field_object.related_model._meta.pk.get_for_dialect(
                    self.DIALECT, "SQL_TYPE"
                ),
                extra=self._table_generate_extra(table=field_object.through),
                comment=self._table_comment_generator(
                    table=field_object.through, comment=field_object.description
                )
                if field_object.description
                else "",
            )
            if not field_object.db_constraint:
                m2m_create_string = m2m_create_string.replace(
                    """,
    ,
    """,
                    "",
                )  # may have better way
            m2m_create_string += self._post_table_hook()
            m2m_tables_for_create.append(m2m_create_string)

        return {
            "table": model._meta.db_table,
            "model": model,
            "table_creation_string": table_create_string,
            "references": references,
            "m2m_tables": m2m_tables_for_create,
        }

    def gen_create_sql(self, model, safe: bool = True) -> str:
        models_to_create = []
        if model._meta.db == self.client:
            model._check()
            models_to_create.append(model)
        tables_to_create = []
        for model in models_to_create:
            tables_to_create.append(self._get_table_sql(model, safe))
        tables_to_create_count = len(tables_to_create)

        created_tables = set()
        ordered_tables_for_create = []
        m2m_tables_to_create = []
        while True:
            if len(created_tables) == tables_to_create_count:
                break
            try:
                next_table_for_create = next(
                    t
                    for t in tables_to_create
                    if t["references"].issubset(created_tables | {t["table"]})
                )
            except StopIteration:
                raise ConfigurationError("Can't create schema due to cyclic fk references")
            tables_to_create.remove(next_table_for_create)
            created_tables.add(next_table_for_create["table"])
            ordered_tables_for_create.append(next_table_for_create["table_creation_string"])
            m2m_tables_to_create += next_table_for_create["m2m_tables"]

        schema_creation_string = "\n".join(ordered_tables_for_create + m2m_tables_to_create)
        return schema_creation_string



from dataclasses import dataclass
from enum import Enum
from types import DynamicClassAttribute

GLOBAL_TZ = 'UTC'
'''全局时区设置'''
GLOBAL_SRV_STATUS = False
'''全局服务状态控制'''


class BaseEnum(Enum):
    """
    继承自该类的枚举设置的模型必须为： name = (value/值, desc/描述或说明) 否则报错

    该类不允许被二次继承
    """

    @DynamicClassAttribute
    def val(self):
        """值"""
        return self.value[0]

    @DynamicClassAttribute
    def label(self):
        """描述"""
        return self.value[1]

    @DynamicClassAttribute
    def data(self):
        """数据(第三项 可选)"""
        if len(self.value) > 2:
            return self.value[2]
        return None

    @classmethod
    def fetch_name(cls, name) -> bool:
        """是否包含命名"""
        return name in cls._member_names_

    @classmethod
    def fetch_val(cls, value):
        """是否包含指定值"""
        v: BaseEnum
        for v in cls._value2member_map_.values():
            if v.val == value:
                return v
        return None

    @classmethod
    def get_name(cls, value):
        """获取指定值的描述"""
        v: BaseEnum
        for v in cls._value2member_map_.values():
            if v.val == value:
                return v.label
        return ""

    @classmethod
    def map_list(cls, outs: list = None):
        """
        以value -- label map的方式返回类型映射

        :param outs: 不输出的值集合, 该参数内的值不做输出
        """
        v: BaseEnum
        if not outs:
            return {v.val: v.label for v in cls._value2member_map_.values()}
        return {v.val: v.label for v in cls._value2member_map_.values() if v.val not in outs}


@dataclass
class Code:
    """状态码对象"""
    val: int
    '''指定响应码'''
    http: int = 200
    '''映射的状态码'''
    msg: str = ''
    '''附加的消息'''


class StaCode:
    """请求的响应状态码"""
    PASS = Code(0, 200, 'Finished.')
    TEND = Code(10, 410, 'Under maintenance.')
    FAIL = Code(11, 400, 'Failed.')
    '''公用请求失败，不确定响应状态请使用该项，并确保响应的文字描述'''
    ERR_ARG = Code(12, 422, 'Invalid params.',)
    '''参数错误或无效'''
    ERR_AUTH = Code(13, 401, 'Invalid authorization.')
    ERR_SIGN = Code(14, 412, 'Invalid sign in.')
    FORBID = Code(15, 403, 'Forbidden')
    NON_PMS = Code(16, 406, 'Non enough permission.')
    EXPIRED = Code(17, 409, 'Request was expired.')
    REAPED = Code(18, 421, 'Request is repeated.')
    ERR_CONF = Code(21, 501, 'Error by configuration.')


class AreaLevel(BaseEnum):
    """地域级别"""
    NATION = 1, '国家'
    PROVINCE = 2, '省/直辖市/地区'
    CITY = 3, '地级市/州/区/县级市'
    COUNTY = 4, '县/区域'
    TOWN = 5, '镇/城区/片区'
    STREET = 6, '街道/村/乡'


class FileSta(BaseEnum):
    NULL = -1, '未设置'
    UPING = 1, '上传中'
    CHECKING = 2, '校验中'
    ERR = 3, '上传错误'
    FINISHED = 4, '完成'



"""
可注入式工具类
可按需要独立注入缓存或日志文件或不重复随机序列等的工具
"""
import decimal
import hashlib
import inspect
import math
from dataclasses import is_dataclass
from datetime import datetime, date
from typing import Union

import aiofiles
import httpx
import orjson


def is_await_fun(func):
    if inspect.isfunction(func) or inspect.ismethod(func):
        return inspect.iscoroutinefunction(func)
    return False


def calc_hash(val: (str, bytes), htype='sha3-256'):
    hash_map = {
        'md5': hashlib.md5,
        'sha256': hashlib.sha256,
        'sha384': hashlib.sha3_384,
        'sha512': hashlib.sha512,
        'sha3-224': hashlib.sha3_224,
        'sha3-256': hashlib.sha3_256,
        'sha3-384': hashlib.sha3_384,
        'sha3-512': hashlib.sha3_512,
    }
    if not isinstance(val, (str, bytes)):
        val = str(val)
    if isinstance(val, str):
        val = val.encode('utf-8')
    hb = hash_map.get(htype) or hashlib.sha3_256
    return hb(val).hexdigest()


def dict_key_to_str(data: dict):
    """将数字型key的字典转化为字符串key类型的字典"""
    if not isinstance(data, dict):
        return data
    new_conf = {}
    for k, v in data.items():
        if isinstance(v, dict):
            v = dict_key_to_str(v)
        elif isinstance(v, list):
            v = [dict_key_to_str(item) for item in v]
        if isinstance(k, (int, float)):
            new_conf[str(k)] = v
        else:
            new_conf[k] = v
    return new_conf


def dict_key_to_int(data: dict):
    """将整数字符串型key的字典转化为整数key类型的字典"""
    if not isinstance(data, dict):
        return data
    new_conf = {}
    for k, v in data.items():
        if isinstance(v, dict):
            v = dict_key_to_str(v)
        elif isinstance(v, list):
            v = [dict_key_to_str(item) for item in v]
        if isinstance(k, str) and k.isdigit():
            new_conf[int(k)] = v
        else:
            new_conf[k] = v
    return new_conf


def json_parse(item_str: Union[str, bytes], log_fun=None):
    """
    序列化json或字典字符串/集合字符串为Python字典/集合数据
    """
    if item_str:
        if isinstance(item_str, (dict, list, tuple)):
            return item_str
        try:
            info = orjson.loads(item_str)
            if isinstance(info, (dict, list, tuple)):
                return info
            return {}
        except (SyntaxError, ValueError, TypeError):
            err_info = f'解析数据出错, 原数据:{item_str}'
            log_fun(err_info) if callable(log_fun) else print(err_info)
    return {}


def json_encode(item, u_byte=False, log_fun=None):
    """
    转换为Json字符串 可以为字典 列表 元组 dataclass等类型
    ps: 使用该方式转换必须保证key不能为数字类型，特别是针对于python字典，有数字类型的key需要先将key转换为字符串
    """
    def dft_set(obj):
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S %z")
        elif isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        elif isinstance(obj, decimal.Decimal):
            return str(obj)
        raise TypeError

    if not item:
        return '' if not u_byte else b''
    option = orjson.OPT_PASSTHROUGH_DATETIME
    if is_dataclass(item):
        option = option | orjson.OPT_SERIALIZE_DATACLASS
    try:
        info = orjson.dumps(item, default=dft_set, option=option)
        return info.decode() if not u_byte else info
    except (SyntaxError, ValueError, TypeError):
        err_info = f'解析数据出错, 原数据:{item}'
        log_fun(err_info) if callable(log_fun) else print(err_info)
        return None


async def read_file(file_path: str, encoding=None, log_fun=None):
    """
    异步读取小文件 大文件请勿使用该方法

    :param file_path: 文件路径(包含文件名)
    :param encoding: (可选)解码模式
    :param log_fun: 日志输出方法
    """
    try:
        async with aiofiles.open(file_path, 'r', encoding=encoding) as f:
            data = await f.read()
            await f.close()
            return data
    except Exception as err:
        err_info = f'读取路径:{file_path} 文件出错:{err}, '
        log_fun(err_info) if callable(log_fun) else print(err_info)
        return None


async def write_file(data, file_path: str, encoding=None, log_fun=None):
    """
    异步写文件

    :param data: 待写入的数据
    :param file_path: 文件路径
    :param encoding: （可选）文件编码格式
    :param log_fun: 日志输出方法
    """
    try:
        async with aiofiles.open(file_path, 'wb', encoding=encoding) as f:
            await f.write(data)
        await f.close()
        return True
    except Exception as err:
        err_info = f'文件写入路径:{file_path} 出错:{err}'
        log_fun(err_info) if callable(log_fun) else print(err_info)
        return False


async def http_get(
        url: str,
        param: dict = None,
        headers: dict = None,
        timeout=10,
        http2=False,
        is_file=False,
        log_fun=None):
    """
    异步http get 请求

    :param url: 请求地址
    :param param: 查询参数
    :param headers: 请求头
    :param timeout: 超时时长 默认10秒
    :param http2: 使用http2.0协议
    :param is_file: 是否为文件
    :param log_fun: 日志输出方法
    """
    async with httpx.AsyncClient(http2=http2) as client:
        if param:
            query_param = ''
            for k, v in param.items():
                query_param += f'{k}={v}&'
            url += f'?{query_param[0: -1]}'
        res = await client.get(url, headers=headers, timeout=timeout)
        if res.status_code < 400:
            if callable(log_fun):
                info = f'Success: {url} \nParams: {param} \nResult: {res.text}'
                log_fun(info)
            return res.text if not is_file else res.content
        err_info = f'Failed: {url} \nParams: {param} \nResult: {res and res.status_code} {res and res.text}'
        log_fun(err_info) if callable(log_fun) else print(err_info)
        return None


async def http_post(
        url: str,
        param: Union[dict, str] = None,
        jsparse=True,
        headers: dict = None,
        timeout=10,
        http2=False,
        log_fun=None):
    """
    异步http post 请求

    :param url: 请求地址
    :param param: 请求参数
    :param jsparse: 是否采用Json格式传输 默认是
    :param headers: 请求头
    :param timeout: 超时时长 默认10秒
    :param http2: 使用http2.0协议
    :param log_fun: 日志输出方法
    """
    async with httpx.AsyncClient(http2=http2) as client:
        if jsparse:
            if not headers:
                headers = {'Content-Type': 'application/json; charset=utf-8'}
            res = await client.post(url, json=param, headers=headers, timeout=timeout)
        else:
            res = await client.post(url, data=param, headers=headers, timeout=timeout)
        if res.status_code < 400:
            if callable(log_fun):
                info = f'Success: {url} \nParams: {param} \nResult: {res.text}'
                log_fun(info)
            return res.text
        err_info = f'Failed: {url} \nParams: {param} \nResult: {res and res.status_code} {res and res.text}'
        log_fun(err_info) if callable(log_fun) else print(err_info)
        return None


def nearby_lonlat(distance: (int, float), lon: float, lat: float):
    """根据经纬度加距离获取矩形坐标范围"""
    radius = 6378137
    r_lon = math.degrees(math.radians(distance / radius * math.cos(math.radians(lon))))
    r_lat = math.degrees(math.radians(distance / radius * math.sin(math.radians(lat))))
    lon_list = (lon + r_lon, lon - r_lon)
    lat_list = (lat + r_lat, lat - r_lat)
    return min(lon_list), max(lon_list), min(lat_list), max(lat_list)



import asyncio
from typing import Union

from nsanic.libs import tool_dt
from nsanic.libs.mult_log import NLogger


class BaseTimed:

    TASK_MAP = {}

    __timed_arr = ('year', 'month', 'day', 'hour', 'minute', 'second')
    __units_map = {'second': 1, 'minute': 60, 'hour': 3600, 'day': 86400}

    def __init__(self, looping=None, run_tz='UTC', log = None):
        self.run_tz = run_tz
        if not looping:
            looping = asyncio.get_event_loop()
        if not log:
            log = NLogger
            log.init_conf()
        self.__run_loop = looping
        self.__logs = log
        self.__run_map = {}

    @property
    def logs(self):
        return self.__logs

    def start(self):
        self.__run_loop and self.__run_loop.run_forever()

    def loginfo(self, *info):
        if not info:
            return
        self.__logs.info(*info) if self.__logs else print(*info)

    def logerr(self, *err):
        if not err:
            return
        self.__logs.error(*err) if self.__logs else print(*err)

    def remove_task(self, name: str):
        tk = self.TASK_MAP.pop(name) if name in self.TASK_MAP else None
        tk and tk.cancel()

    def add_loop_task(self, name: str, interval: int, fun, fun_param: Union[list, tuple] = None, delay=0):
        """
        添加循环任务
        :param name: 任务名称
        :param interval: 运行间隔
        :param fun: 协程或可等待的执行函数
        :param fun_param: 执行函数参数
        :param delay: 启动后延时运行时间，0不运行
        :return:
        """

        async def loop_fun():
            delay and ((await fun(*fun_param)) if asyncio.iscoroutinefunction(fun) else fun(*fun_param))
            while 1:
                interval and (await asyncio.sleep(interval))
                try:
                    (await fun(*fun_param)) if asyncio.iscoroutinefunction(fun) else fun(*fun_param)
                except Exception as err:
                    self.loginfo('循环任务运行出错', err)

        if interval and interval <= 2:
            self.loginfo('存在间隔小于2秒的任务, 将不加入循环任务执行')
        if not fun_param:
            fun_param = []
        self.remove_task(name)
        tk = self.__run_loop.create_task(loop_fun(), name=name)
        self.TASK_MAP.update({name: tk})

    def add_timed_task(
            self,
            name: str,
            timed: Union[list, tuple],
            fun,
            fun_param: Union[list, tuple] = None,
            run_now=False):
        """
        添加定时任务
        :param name: 任务名称 重复的将会覆盖
        :param timed: 定时配置
        :param fun: 定时方法
        :param fun_param: 任务执行参数
        :param run_now: 是否立即执行一次
        """
        async def timed_fun(_timed, _run_now):
            if _run_now:
                ((await fun(*fun_param)) if asyncio.iscoroutinefunction(fun) else fun(*fun_param))
                self.__run_map[name] = tool_dt.cur_dt(tz=self.run_tz)
            while 1:
                g_sta, nt, over_sta = self.__get_next_datetime(_timed)
                if not nt:
                    break
                try:
                    self.__check_run(name, g_sta, timed) and (
                        await fun(*fun_param)) if asyncio.iscoroutinefunction(fun) else fun(*fun_param)
                except Exception as err:
                    self.loginfo('循环任务运行出错', err)
                dt = tool_dt.cur_dt(tz=self.run_tz)
                if dt > nt:
                    self.logs.info(
                        f'当前定时任务执行超时，将跳过当前周期进入下一周期，超时时长：{(dt - nt).total_seconds()}')
                    _, nt, _ = self.__get_next_datetime(timed, dt)
                    if not nt:
                        break
                interval = round((nt - dt).total_seconds(), 4)
                self.logs.info(f'定时任务下次运行等待时长:{interval}', timed, g_sta)
                await asyncio.sleep(interval)
        self.check_timed(timed)
        timed = list(timed)
        if not fun_param:
            fun_param = []
        self.remove_task(name)
        tk = self.__run_loop.create_task(timed_fun(timed, run_now), name=name)
        self.TASK_MAP.update({name: tk})

    @classmethod
    def check_timed(cls, timed: Union[list, tuple]):
        if not timed:
            raise Exception(f'{cls.__class__.__name__}未指定定时运行参数或间隔运行参数')
        if not isinstance(timed, (tuple, list)):
            raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，要求必须是(年 月 日 时 分 秒)')
        if len(timed) != 6:
            raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，要求必须是(年 月 日 时 分 秒)')
        for val in timed:
            if (val != '*') and (not str(val).isdigit()):
                raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，必须是字符型数字或*(不指定)')
        if ''.join(list(map(str, timed))) == ('*' * 6):
            raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，不能全部指定为*')
        if (timed[0] != '*') and (int(timed[0]) < 2023):
            raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，年份不能小于当年')
        if (timed[1] != '*') and ((int(timed[1]) < 0) or (int(timed[1]) > 12)):
            raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，月份不符合要求')
        if (timed[2] != '*') and ((int(timed[2]) < 0) or (int(timed[2]) > 31)):
            raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，日期不符合要求')
        if (timed[3] != '*') and ((int(timed[3]) < 0) or (int(timed[3]) >= 24)):
            raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，小时不符合要求')
        if (timed[4] != '*') and ((int(timed[4]) < 0) or (int(timed[4]) > 60)):
            raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，分钟不符合要求')
        if (timed[5] != '*') and ((int(timed[5]) < 0) or (int(timed[5]) > 60)):
            raise Exception(f'{cls.__class__.__name__}定时参数不符合规范，秒不符合要求')

    def __check_run(self, name: str, sta_list, timed):
        """确保睡眠周期内每天执行一次"""
        if all(sta_list):
            return True
        m_len = timed.count('*')
        new_list = [sta_list[i] for i in range(m_len + 1)]
        if (0 < m_len < 5) and all(new_list):
            last_dt = self.__run_map.get(name)
            if not last_dt:
                return True
            cur_dt = tool_dt.cur_dt(tz=self.run_tz)
            if getattr(cur_dt, self.__timed_arr[m_len - 1]) != getattr(last_dt, self.__timed_arr[m_len - 1]):
                self.__run_map[name] = cur_dt
                return True
        return False

    def __get_next_datetime(self, timed, cur_dt=None):
        gt_sta = [False] * 6
        if not cur_dt:
            cur_dt = tool_dt.cur_dt(tz=self.run_tz)
        dt_list = [cur_dt.year, cur_dt.month, cur_dt.day, cur_dt.hour, cur_dt.minute, cur_dt.second]
        index_nt = -1
        for index, val in enumerate(timed):
            if val != '*':
                if int(val) == dt_list[index]:
                    gt_sta[index] = True
            else:
                gt_sta[index] = True
                index_nt = index
        if index_nt == -1:
            return gt_sta, None, False
        nt_list = [dt_list[i] for i in range(index_nt + 1)] + list(map(int, timed[index_nt + 1:]))
        next_dt = tool_dt.create_dt(*nt_list, tz=self.run_tz)
        over_sta = False
        if next_dt <= cur_dt:
            next_dt = self.cult_second_dt(next_dt, self.__timed_arr[index_nt])
            over_sta = True
        return gt_sta, next_dt, over_sta

    def cult_second_dt(self, dt, add_str: str, add_count: int = 1):
        if add_str in self.__units_map:
            return tool_dt.to_datetime(int(dt.timestamp()) + (add_count * self.__units_map[add_str]), tz=self.run_tz)
        if add_str == 'month':
            days_arr, year, month = [], dt.year, dt.month
            for i in range(add_count):
                month += 1
                if month <= 12:
                    days_arr.append(tool_dt.month_days(year, month))
                else:
                    year += 1
                    month = month - 12
                    days_arr.append(tool_dt.month_days(year, month))
            total_sec = sum(days_arr) * 86400
            return tool_dt.to_datetime(int(dt.timestamp()) + total_sec, tz=self.run_tz)
        total_sec = sum([(365 if not tool_dt.check_leap(dt.year + i) else 366) for i in range(add_count)]) * 86400
        return tool_dt.to_datetime(int(dt.timestamp()) + total_sec, tz=self.run_tz)



from setuptools import find_packages, setup

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='nsanic',
    version='2025.04.01',
    author='DHDONG',
    author_email='zscych@qq.com',
    license='MIT',
    packages=find_packages(),

    description='A web framework who is use sanic + tortoise-orm + redis to quick create http&websocket server.'
                '\nYou can quickly build a new project by command -- sanicGuider',
    long_description=long_description,
    long_description_content_type='text/markdown',

    install_requires=[
        'sanic >= 22.12.0',
        'redis[hiredis] >= 4.6.0',
        'pyjwt >= 0.2.6',
        'httpx >= 0.23.0',
        'orjson >= 3.9.15',
        # 'pycryptodome >= 3.16.0',
    ],
    entry_points={
        "console_scripts": [
            "sanicGuider=nsanic:mult_creator",
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    python_requires='>=3.9,<3.12',
    keywords=['WebServer', 'Sanic', 'WebSite Develop', 'Web Framework']
)



from typing import Union

from sanic import Blueprint

from nsanic.handler_http import Urls


class BaseBlue:

    DEFAULT_APIS: (list, tuple) = []
    bpo: Blueprint = None
    name: str = None
    version: Union[str, int, float] = None
    url_prefix: str = None
    ws_handlers = []

    @classmethod
    def init(cls, srv_name: str, bp_name: str = None, version: Union[str, int, float]=None):
        if not bp_name:
            bp_name = cls.__name__
            len_name = len(bp_name)
            bp_name = f'{bp_name[0].lower()}{bp_name[1: len_name]}'
            if bp_name[-2: len_name].lower() == 'bp':
                bp_name = bp_name[0: -2]
        cls.name = bp_name
        cls.version = version
        srv_arr = srv_name.split('_')
        srv_str = f"{srv_arr[0]}{''.join([srv_arr[i].capitalize() for i in range(1, len(srv_arr))])}"
        new_prefix = srv_str if cls.name == 'main' else f'{srv_str}/{cls.name}'
        cls.url_prefix = new_prefix
        cls.bpo = Blueprint(name=cls.name, url_prefix=new_prefix, version=cls.version, version_prefix='/v')

    @classmethod
    def load_default_api(cls):
        """加载默认API"""
        cls.DEFAULT_APIS and cls.__load_api(cls.DEFAULT_APIS)

    @classmethod
    def __load_api(cls, item_list):
        for item in item_list:
            if not item:
                continue
            if not isinstance(item, Urls):
                raise Exception(f'Invalid router urls object: {item}')
            if not item.handler:
                raise Exception(f'Invalid router handler object: {item}')
            if not item.router:
                if hasattr(item.handler, 'as_view') or hasattr(item.handler, 'wsrouter'):
                    item = [f'/{cls.name}', item]
                else:
                    continue
            version = cls.version if not item.ver else item.ver
            if item.router and item.router[0] != '/':
                rpath = f'/{item.router}'
            else:
                rpath = item.router
            if hasattr(item.handler, 'as_view'):
                cls.bpo.add_route(item.handler.as_view(), rpath, version=version, name=item.name or None)
                continue
            if hasattr(item.handler, 'wsrouter'):
                ws_obj = item.handler.init()
                cls.bpo.add_websocket_route(ws_obj.wsrouter, rpath, version=version, name=item.name or None)
                cls.ws_handlers.append(ws_obj)

    @classmethod
    def load_apis(cls, item_list: list):
        """
        批量注册路由
        :param item_list: 加载模型必须为[(路由,API处理器类,(可选)版本号)]
        """
        cls.__load_api(item_list)

    @classmethod
    def register_api(cls, url: str, handler, ver: Union[int, str, float] = None):
        """注册单路由"""
        cls.bpo and cls.bpo.add_route(handler, url, ver)

    @classmethod
    def add_signal(cls, event: str, handler, condition=None, exclusive=True):
        """
        注册信号

        """
        cls.bpo and cls.bpo.add_signal(handler, event, condition=condition, exclusive=exclusive)

    # def add_bp(self):
    #     """注册其它蓝图组路由"""
    #     self.__blue_obj.group()



from sanic import Request, HTTPResponse

from nsanic.libs import consts
from nsanic.libs.component import ConfMeta
from nsanic.libs.manager import HeaderSet


class CorsMiddle(ConfMeta):

    @classmethod
    async def main(cls, req: Request):
        """通用跨域适配"""
        headers = HeaderSet.out(cls.conf)
        if req.method == 'OPTIONS':
            if ('*' not in cls.conf.ALLOW_ORIGIN) and (req.server_name not in cls.conf.ALLOW_ORIGIN):
                return HTTPResponse('', status=403, headers=headers)
            return HTTPResponse('', status=204, headers=headers)
        if not consts.GLOBAL_SRV_STATUS:
            return HTTPResponse('Server is not running', status=503, headers=headers)



from nsanic.libs.mult_log import NLogger


class LogMeta:

    @classmethod
    def logerr(cls, *err: str):
        NLogger.error(*err)

    @classmethod
    def loginfo(cls, *info):
        NLogger.info(*info)


class ConfMeta(LogMeta):

    conf = None

    @classmethod
    def set_conf(cls, conf):
        cls.conf = conf
